classdef lr_newt < handle
%LR_NEWT Solve the Newton subproblem
%       AP(d)A' * y = b
% for LP / SDP in SeDuMi using a low-rank preconditioner and PCG
% Author: Richard Y. Zhang (ryz@berkeley.edu)
properties(SetAccess=private)
    rank; rp; fill;
    pastCGIter = [0,0,0]; 
end
properties(Access=private)
    % Problem info
    Alp; % Lp blocks
    Asdp; Vsym; % SDP blocks and symmetric bases
    K; nn; m;
    
    % Factorization for the preconditioner
    L; D; perm;
    
    % Scaling matrices
    dl; % LP
    Ws; Vs; ds; % SDP
    beta; kappasqrt;
end
properties(Constant)
    % Krylov solver parameters
    solver = 1; 
    max_iter = 1000;
    restart = 25;
    
    % Debug mode plots the eigenvalues of W
    debug = true; 
    
    % Heuristic for figuring out the solution rank k
    dlogkappa_thresh = 1;
    max_rank = 10;
end
methods
%==========================================================================
% Initialization
% DATA MUST BE SUPPLIED IN SEDUMI FORMAT
%==========================================================================
function [ p ] = lr_newt(A, K)
    [n2,m] = size(A);
    % We like A to be a column-major matrix (with more rows than cols)
    if n2 < m 
        A = A.';
    end
    blknum = numel(K.s);
    
    %----------------------------------------------------------------------
    % Resymmetricize the data for numerical stability
    %----------------------------------------------------------------------
    % LP part
    Alp = A(1:K.l,:); 
    
    % Symmetricize each SDP block in A
    % Reduce each Ai to its n(n+1)/2 dofs (instead of n^2 elements)
    [Asdp,Vsym] = deal(cell(blknum,1));
    for i = 1:numel(K.s)
        % Get this matrix
        Vsym{i} = symbasis(K.s(i));
        idx = K.sblkstart(i):(K.sblkstart(i+1)-1);
        Asdp{i} = Vsym{i}' * A(idx,:);
    end
    
    % Store
    p.Alp = Alp;
    p.Asdp = Asdp; 
    p.Vsym = Vsym;
    p.K = K;
    p.m = size(A,2);
end

function set_scale(p, d)
    K = p.K; % Reduce clutter
    %----------------------------------------------------------------------
    % Generate the scaling matrices
    %----------------------------------------------------------------------
    blknum = numel(K.s);
    [Vs, ds, Ws] = deal(cell(blknum,1));
    % These deal with sedumi's internal indexing system.
    % cf. psdscale.m, lines 59 & 63
    startindices=K.sblkstart-K.mainblks(end)+1; 
    permindices=cumsum([1; K.s(:)]); 
    for i = 1:blknum
        % Get lower triangular Cholesky factor
        %       Ws(perm, perm) = Ls * Ls'
        % cf. psdscale.m, line 78 & 81
        Ls = tril(reshape(d.u(startindices(i):startindices(i+1)-1), K.s(i), K.s(i)));
        if isfield(d, 'perm') && ~isempty(d.perm)
            perm = d.perm(permindices(i):permindices(i+1)-1);
        else
            perm = 1:K.s(i);
        end
        
        % Output the scaling matrix
        [U, S, ~] = svd(Ls);
        Vs{i}(perm, :) = U;
        ds{i} = diag(S).^2;
        
        % We from Ws from Vs and ds to make them match up to machine
        % precision
        Ws{i} = bsxfun(@times, Vs{i}, sqrt(ds{i}(:)).');
        Ws{i} = Ws{i} * Ws{i}';
    end
    [p.dl, p.Ws, p.Vs, p.ds] = deal(d.l,Ws,Vs,ds);
    
    %----------------------------------------------------------------------
    % Initialize the preconditioner
    %----------------------------------------------------------------------
    % Form LP part of Hessian directly
    % (We pretend that there are no dense columns....)
    ada = p.Alp' * diag(sparse(p.dl)) * p.Alp;
        
    % Construct the preconditioner one SDP block at a time
    [AtU, Z] = deal(cell(1,blknum));
    rp = zeros(blknum,1); % Prec rank of each block
    for i = 1:blknum        
        % Set up preconditioner          
        % Look for the logarithmic median and the biggest outliers
        n = K.s(i); ev = ds{i}(:); logev = log(ev);
        tau = exp(median(logev));
        
        % Approximate this matrix by its logmedian
        ada = ada + tau^2 * (p.Asdp{i}' * p.Asdp{i});
        
        % Heuristic for finding the rank k
        max_rank = floor(min(p.max_rank, n/2));
        dlogkappa = logev(2:max_rank+1)-logev(1:max_rank);
        k = find(dlogkappa < -p.dlogkappa_thresh, 1, 'last');
        if isempty(k), k = 0; end
        
        if p.debug 
            % Plot the eigenvalues of the scaling matrix
            figure(1); clf
            hold on
            plot(1:n, ds{i}, '-k','linewidth',2);
            plot((1:n-1), ds{i}(1:end-1)./ds{i}(2:end), '--k');
            hold off
            xlim([1, n]);
            set(gca, 'xscale', 'log', 'yscale', 'log');
            title(sprintf('k = %d', i));
            legend('eig(W)', 'ratio');
            grid on;
            pause(0.001);
        end
        
        % Low-rank perturbation?
        if k == 0
            rp(i) = 0;
        else            
            % Split W = W0 + U1 Z1 U1' + U2 Z2 U2'
            % The remaining eigenvalues are approximated by the smallest
            % eigenvalue
            dw = ev(1:k) - tau; % Always positive by construction
            U = bsxfun(@times, Vs{i}(:, 1:k), sqrt(dw)');
            AtU{i} = tau^(1.5) * p.Asdp{i}' * (p.Vsym{i}' * kron(speye(n), U));
            Z{i} = tau^2 / 2 * speye(n * k);
            rp(i) = k;
        end
    end
    
    % Combine the preconditioner and factorize
    rk = rp(:)'*K.s(:);
    AtU = [AtU{:}]; Z = blkdiag(Z{:});
    AugSys = [ada,   AtU;
              AtU',  -Z];
    [p.L, p.D, p.perm] = ldl(AugSys, 0, 'vector');
    p.D = diag(p.D);
    [p.rp, p.rank] = deal(rp, rk);
    p.fill = (2*nnz(p.L) - length(p.L)) / nnz(AugSys);
end

%==========================================================================
% Solution
%==========================================================================
function [y, iters, r] = solve(p, b, tol)
% Apply the preconditioner while fully exploiting the kronecker structure. 
% base case: no preconditioner needed
    % Solve this using my_pcg
    [y, flag, normr, iters, resvec2, r] = ...
    my_pcg(@p.mvhess, b, tol, p.max_iter, @p.prec);

    % Debug mode checks the residual and plots the residual convergence
    % curve if convergence fails. 
    if p.debug && normr >= tol
        warning(['PCG failed to reach the desired tolerance (flag = %d)\n'...
            'norm(r,inf) = %e, tol = %e'], flag, normr, tol);
        figure(2); semilogy(resvec2);
        pause(0.001);
    end
    
    % Store the number of iterations
    p.pastCGIter(1:end-1) = p.pastCGIter(2:end);
    p.pastCGIter(end) = iters;
end
end
%==========================================================================
% Private preconditioner functions
%==========================================================================
methods(Access=private)
function y = prec(p, y)
    % Apply the preconditioner while fully exploiting the kronecker structure. 
    y = [y; zeros(p.rank, size(y,2))];
    y = p.L \ y(p.perm,:);
    y = p.D .\ y;
    y = p.L' \ y;
    y(p.perm,:) = y;
    y = y(1:p.m,:);
end
function x = mvhess(p, y)
% Implicitly compute y = AP(d)A' * y
    % Linear scaling
    x = p.dl .* (p.Alp*y);
    x = (x'*p.Alp)';

    % SDP scaling 
    for i = 1:numel(p.K.s)
        % Get matrix
        X = p.Vsym{i}*(p.Asdp{i}*y);
        X = reshape(X, p.K.s(i), p.K.s(i)); % retrieve matrix
        
        % Scale this matrix
        X = p.Ws{i} * X * p.Ws{i}; % Note that X is forced to be symmetric
        
        % Add contribution
        x = x + ((X(:)' * p.Vsym{i}) * p.Asdp{i})';
    end
end
end
end

function [x,flag,absres,ii,resvec,r] = my_pcg(A,b,abs_inf_tol,maxit,M1)
%MY_PCG Preconditioned Conjugate Gradients Method.
%      Modifications:
%      -- Termination criterion is now residual abs. inf. norm.
%      -- Removed stagnation detection, which gives up too early around the
%         numerical floor.
%      -- "iter" is the total number of iterations, not the best.
%      -- Removed support for matrix input, missing arguments
%      (Simplified and significantly modified from the MATLAB internal code)
n = numel(b);

% Provided a preconditioner?
existM1 = nargin > 4 && ~isempty(M1);

% Check for all zero right hand side vector => all zero solution
n2b = norm(b, inf);                % Norm of rhs vector, b
if (n2b == 0)                      % if    rhs vector is all zeros
    x = zeros(n,1);                % then  solution is all zeros
    flag = 0;                      % a valid solution has been obtained
    absres = 0;                    % the relative residual is actually 0/0
    ii = 0;                        % no iterations need be performed
    resvec = 0;                    % resvec(1) = norm(b-A*x) = norm(0)
    r = zeros(n,1);
    return
end

% Set up for the method
x = zeros(n,1);
flag = 1;
xmin = x;                          % Iterate which has minimal residual so far
r = b - A(x);
normr = norm(r, inf);              % Norm of residual
normr_act = normr;
if (normr <= abs_inf_tol)                 % Initial guess is a good enough solution
    flag = 0;
    absres = normr;
    ii = 0;
    resvec = normr;
    return
end

resvec = zeros(maxit+1,1);         % Preallocate vector for norm of residuals
resvec(1,:) = normr;               % resvec(1) = norm(b-A*x0)
normrmin = normr;                  % Norm of minimum residual
rho = 1;

% loop over maxit iterations (unless convergence or failure)
for ii = 1 : maxit
    if existM1 % Precondition?
        z = M1(r);
        if ~all(isfinite(z))
            flag = 2;
            break
        end
    else % no preconditioner
        z = r;
    end
    
    rho1 = rho;
    rho = r' * z;
    if ((rho == 0) || isinf(rho))
        flag = 4;
        break
    end
    if (ii == 1)
        p = z;
    else
        beta = rho / rho1;
        if ((beta == 0) || isinf(beta))
            flag = 4;
            break
        end
        p = z + beta * p;
    end
    q = A(p);
    pq = p' * q;
    if ((pq <= 0) || isinf(pq))
        flag = 4;
        break
    else
        alpha = rho / pq;
    end
    if isinf(alpha)
        flag = 4;
        break
    end    
    x = x + alpha * p;             % form new iterate
    r = r - alpha * q;
    normr = norm(r, inf);
    normr_act = normr;
    resvec(ii+1,1) = normr;
    
    % check for convergence
    if normr <= abs_inf_tol
        r = b - A(x);
        normr_act = norm(r, inf);
        resvec(ii+1,1) = normr_act;
        if (normr_act <= abs_inf_tol)
            flag = 0;
            break
        end
    end
    if (normr_act < normrmin)      % update minimal norm quantities
        normrmin = normr_act;
        xmin = x;
    end
end                                % for ii = 1 : maxit

% returned solution is first with minimal residual
if (flag == 0)
    absres = normr_act;
else
    r_comp = b - A(xmin);
    if norm(r_comp, inf) <= normr_act
        x = xmin;
        absres = norm(r_comp, inf);
    else
        absres = normr_act;
    end
end

% truncate the zeros from resvec
if ((flag <= 1) || (flag == 3))
    resvec = resvec(1:ii+1,:);
else
    resvec = resvec(1:ii,:);
end

end

function [V] = symbasis(n)
%SYMBASIS Basis for the set of real symmetric matrices within the linear 
% space of all n x n matrices.
ndof = n*(n+1)/2;

% Enumerate the subscripts of a lower triangular matrix
[ii,jj] = find(sparse(tril(ones(n))));

% Convert the subscrips into indices
ii1 = ii + (jj-1)*n;
ii2 = (ii-1)*n + jj;

% Number according to DOFs
jj = (1:ndof)';

% Scaling to make the matrix unitary
kk = sqrt(0.5)*ones(ndof,2);
kk(ii1==ii2,:) = 0.5*ones(n,2);

% Output
V = sparse([ii1,ii2],[jj,jj],kk,n^2,ndof);
end